
(* rm -rf d, if it exists *)
fun removeDir d =
    (let open OS.FileSys
     in if isDir d handle _ => false then
	    (let val stream = openDir d
	     in while case readDir stream of
			  SOME f => (remove (OS.Path.joinDirFile {dir=d,file=f}); true)
			| _ => false do ();
		rmDir d
	     end
	     handle _ => (print ("cannot remove directory "^d^"\n"); raise BYE))
	else if (TextIO.openIn d; true) handle _ => false then
	    (remove d handle _ => (print ("cannot remove file "^d^"\n"); raise BYE))
	else
	    ()
     end)
    handle _ => ();
 

(* main *)
fun top () =
	let val args = CommandLine.arguments ()
	    val () = command args
	in (* check/set error stream *)
	   case !ErrPrint.name of
		  "" => ()
		| f  => let val f = OS.Path.mkCanonical f
			in ErrPrint.FILE := TextIO.openOut f handle x => (print ("cannot open "^f^"\n"); raise BYE)
			end;
	   (* check/set input stream *)
	   case !Input.name of
		  "" => ()
		| f  =>	let val f = OS.Path.mkCanonical f
			in case OS.Path.ext f of
				  SOME e => if mem e ["fcr"] then () else (print ("unknown format for input file: "^f^"\n"); raise BYE)
				| e      => (print ("unknown format for input file: "^f^"\n"); raise BYE);
		   	   Input.FILE := TextIO.openIn f handle x => (print ("cannot open "^f^"\n"); raise BYE)
			end;
	   (* check/set output stream *)
	   case !OutPrint.name of
		  "" => (case !OUTPUT of
                                 TTS => (print ("cannot create tts on standard output, please specify outfile\n"); raise BYE)
                               | HIPPO => if HPX then
					      (print ("cannot create hippo on standard output, please specify outfile\n"); raise BYE)
					  else
					      ()
                               | _ => ())
		| f  =>	let val f = OS.Path.mkCanonical (strip f)
				    val e = case OS.Path.ext f of SOME x => x | NONE => ""
			in case !OUTPUT of
			       Unknown => (* output defined by outfile type *)
			       (case e of
				    "fcr" => (OUTPUT := Fiacre;
                                              OutPrint.FILE := TextIO.openOut f handle x => (print ("cannot open "^f^"\n"); raise BYE))
                                  | "tts" => (OUTPUT := TTS;
                                              (* create f dir; files will be created by printer *)
                                              removeDir f;      (* rm -rf f *)
                                              OS.FileSys.mkDir f handle x => (print ("cannot create "^f^"\n"); raise BYE))
                                  | "hippo" => if HPX then
						   (OUTPUT := HIPPO; HIPPOEXT := true; PROPERTIES := false;
						    (* create f dir; files will be created by printer *)
						    removeDir f;      (* rm -rf f *)
						    OS.FileSys.mkDir f handle x => (print ("cannot create "^f^"\n"); raise BYE))
					       else
						   (print ("hippo code generator not provided by this copy of frac\n"); raise BYE)
				  | "lotos" => (print ("cannot generate Lotos code, consider flac\n"); raise BYE)
				  | _ => (print ("unknown format for output file: "^f^"\n"); raise BYE))
			     | NoOutput => (* case -p or -c, supersedes outfile type, but warn if confliting *)
			       (if mem e ["fcr", "tts", "hippo", "lotos"] then
				    (print "output and outfile type are incompatible\n"; raise BYE)
				else
				    ();
				OutPrint.FILE := TextIO.openOut f handle x => (print ("cannot open "^f^"\n"); raise BYE))
			     | Fiacre =>
			       (if mem e ["tts", "hippo", "lotos"] then
				    (print "output and outfile type are incompatible\n"; raise BYE)
				else
				    ();
				OutPrint.FILE := TextIO.openOut f handle x => (print ("cannot open "^f^"\n"); raise BYE))
			     | Abstract =>
			       (if mem e ["fcr", "tts", "hippo", "lotos"] then
				    (print "output and outfile type are incompatible\n"; raise BYE)
				else
				    ();
				OutPrint.FILE := TextIO.openOut f handle x => (print ("cannot open "^f^"\n"); raise BYE))
			     | TTS =>
			       (if mem e ["fcr", "hippo", "lotos"] then
				    (print "output and outfile type are incompatible\n"; raise BYE)
				else
				    ();
				removeDir f;      (* rm -rf f *)
				OS.FileSys.mkDir f handle x => (print ("cannot create "^f^"\n"); raise BYE))
			     | HIPPO =>
			       if HPX then
				   (if mem e ["fcr", "tts", "lotos"] then
					(print "output and outfile type are incompatible\n"; raise BYE)
				    else
					();
				    removeDir f;      (* rm -rf f *)
				    OS.FileSys.mkDir f handle x => (print ("cannot create "^f^"\n"); raise BYE))
			       else
				   ()
			     | LOTOS => (print ("cannot generate Lotos code, consider flac\n"); raise BYE)
			end;
	   (* go *)
	   Type.SHVREAD := true;    (* shared vars can be read anywhere    *)
	   Type.SHVWRIT := true;    (* shared vars can be written anywhere *)
	   let val (p,p') = Type.typecheck (Parse.parse ())
	   in if !STRIP then PROPERTIES := false else ();
	      Init.initcheck p';
              case !OUTPUT of
		  NoOutput => ()
		| Fiacre => Print.print (if !XPAND then Observers.expandDecls p' else
					 if !TRANSF then p' else p)
		| _ => Compile.compile p'
	   end;
	   (* exit *)
	   ErrPrint.flush (); OutPrint.flush (); OS.Process.exit OS.Process.success
	end
	handle BYE => (ErrPrint.flush (); OutPrint.flush (); OS.Process.exit OS.Process.failure);


(* making the application (for mlton) ----------------------------------- *)

mkApp top "frac";

